==> pan-0.90.tar.bz2: pan-0.90/pan/tasks/task-article.cc <==
/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <cassert>
#include <ostream>
#include <sstream>
#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include <pan/general/i18n.h>
#include <pan/general/log.h>
#include <pan/usenet-utils/decode.h>
#include <pan/usenet-utils/mime-utils.h>
#include <pan/data/article-cache.h>
#include "task-article.h"

using namespace pan;

/***
****
***/

namespace
{
  std::string get_description (const Article& article, bool save)
  {
    std::string stripped;
    mime::remove_multipart_from_subject (article.subject.c_str(), stripped);

    char buf[1024];
    if (save)
      std::snprintf (buf, sizeof(buf), _("Saving %s"), stripped.c_str());
    else
      std::snprintf (buf, sizeof(buf), _("Reading %s"), stripped.c_str());
    return std::string (buf);
  }
}

TaskArticle :: TaskArticle (const GroupServer         & group_server,
                            const Article             & article,
                            ArticleCache              & cache,
                            ArticleRead               & read,
                            Task::Listener            * listener,
                            const StringView          & save_path):
  Task (save_path.empty() ? "BODIES" : "SAVE", get_description (article, !save_path.empty())),
  _save_path (save_path.str, save_path.len),
  _cache (cache),
  _read (read),
  _article (article),
  _time_posted (article.time_posted),
  _finished_proc_has_run (false),
  _save_when_finished (!save_path.empty())
{
  cache.reserve (article.parts);

  if (listener != 0)
    add_listener (listener);

  // build a list of all the parts we need to download.
  // also calculate need_bytes and all_bytes for our Progress status.
  unsigned long need_bytes(0), all_bytes(0);
  foreach_const (Article::parts_t, article.parts, it)
  {
    all_bytes += it->bytes;
    if (!cache.contains (it->message_id))
    {
      need_bytes += it->bytes;

      Needed n;
      n.part = *it;
      foreach_const (Xref, article.xref, xit) {
        quarks_t servers;
        group_server.group_get_servers (xit->group, servers);
        foreach_const (quarks_t, servers, sit)
          n.xref.insert (*sit, xit->group, 0ul);
      }
      _needed.push_back (n);
    }
  }

  // initialize our progress status...
  init_steps (all_bytes);
  set_step (all_bytes - need_bytes);
  if (save_path.empty())
    set_status (article.subject.c_str());
  else
    set_status_va (_("Saving %s"), article.subject.c_str());
 
  update_work ();
}

TaskArticle :: ~TaskArticle ()
{
  _cache.release (_article.parts);
}

void
TaskArticle :: update_work ()
{
  // which servers could we use right now?
  int working (0);
  quarks_t servers;
  foreach_const (needed_t, _needed, it)
    if (it->nntp)
      ++working;
    else foreach_const (Xref, it->xref, xit)
      servers.insert (xit->server);

  if (!servers.empty())
    _state.set_need_nntp (servers);
  else if (working)
    _state.set_working ();
  else {
    _state.set_completed ();
    set_finished (OK);
  }

  if (_state._work == COMPLETED && !_finished_proc_has_run) {
    _finished_proc_has_run = true;
    on_finished ();
  }
}

/***
****
***/

void
TaskArticle :: use_nntp (NNTP * nntp)
{
  // find which part, if any, can use this nntp
  Needed * needed (0);
  for (needed_t::iterator it(_needed.begin()), end(_needed.end()); !needed && it!=end; ++it)
    if (it->nntp==0 && it->xref.has_server(nntp->_server))
      needed = &*it;

  if (!needed)
  {
    // couldn't find any work for this nntp...
    // maybe our needs-work list is out of date, so call update_work()
    // before giving the nntp back.
    update_work ();
    check_in (nntp, true);
  }
  else
  {
    needed->nntp = nntp;
    needed->buf.clear ();
    needed->buf.reserve (needed->part.bytes);

    Quark group;
    unsigned long unused;
    needed->xref.find (nntp->_server, group, unused);
    nntp->article (group, needed->part.message_id.c_str(), this);
    update_work ();
  }
}

/***
****
***/

void
TaskArticle :: on_nntp_line  (NNTP               * nntp,
                             const StringView   & line)
{
  // FIXME: ugh, this is called for _every line_...
  Needed * needed (0);
  foreach (needed_t, _needed, it) {
    if (it->nntp == nntp) {
      needed = &*it;
      break;
    }
  }
  assert (needed);

  Needed::buf_t& buf (needed->buf);
  buf.insert (buf.end(), line.begin(), line.end());
  buf.insert (buf.end(), '\n');
  increment_step (line.len);
}

void
TaskArticle :: on_nntp_done  (NNTP    * nntp,
                              Health    health)
{
  // find the Needed using this nntp...
  needed_t::iterator it;
  for (it=_needed.begin(); it!=_needed.end(); ++it)
    if (it->nntp == nntp)
      break;
  g_assert (it != _needed.end());

  if (health == OK) { // if download succeeded, save it in the cache
    const StringView view (&it->buf.front(), it->buf.size());
    _cache.add (it->part.message_id, view);
    ++_stats[nntp->_server];
  }

  if (health == FAIL) // if server doesn't have that article...
    it->xref.remove_server (nntp->_server);

  // if we got it or still have other options, we're okay
  _state.set_health (health==FAIL && it->xref.empty() ? FAIL : OK);

  if (health==OK)
    _needed.erase (it);
  else
    it->nntp = 0;

  update_work ();
  check_in (nntp, health!=RETRY);
}

void
TaskArticle :: on_finished ()
{
  // this is mostly interesting for making sure multiserver is distributing the load right...
  // so this log message will be removed before the stable release.
  // That's why it's not marked for translation
  if (_stats.size() > 1) {
    std::ostringstream o;
    unsigned int total (0);
    foreach_const (stats_t, _stats, it) total += it->second;
    o << "Fetched " << _article.subject << " in " << total << " parts: ";
    foreach_const (stats_t, _stats, it)
      o << it->second << " from " << it->first << " ";
    Log::add_info (o.str().c_str());
  }

  if (_save_when_finished)
  {
    ArticleCache :: mid_sequence_t mids;
    foreach_const (Article::parts_t, _article.parts, it)
      mids.push_back (it->message_id);

    GMimeMessage * mm = _cache.get_message (_article.parts);
    if (mm) {
      Decode :: decode (_save_path,
                        Quark(), // FIXME: fallback filename,
                        mm,
                        *this);
      _read.mark_read (_article);
      g_object_unref (G_OBJECT (mm));
    }
  }
}

unsigned long
TaskArticle :: get_bytes_remaining () const
{
  unsigned long bytes (0);
  foreach_const (needed_t, _needed, it) // parts not fetched yet...
    bytes += (it->part.bytes - it->buf.size());
  return bytes;
}

==> pan-0.95.tar.bz2: pan-0.95/pan/tasks/task-article.cc <==
/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <cassert>
#include <ostream>
#include <sstream>
extern "C" {
  #define PROTOTYPES
  #include <uulib/uudeview.h>
};
#include <pan/general/debug.h>
#include <pan/general/file-util.h>
#include <pan/general/foreach.h>
#include <pan/general/i18n.h>
#include <pan/general/log.h>
#include <pan/usenet-utils/decode.h>
#include <pan/usenet-utils/mime-utils.h>
#include <pan/data/article-cache.h>
#include "task-article.h"

using namespace pan;

/***
****
***/

namespace
{
  std::string get_description (const Article& article, bool save)
  {
    std::string stripped;
    mime::remove_multipart_from_subject (article.subject.c_str(), stripped);

    char buf[1024];
    if (save)
      std::snprintf (buf, sizeof(buf), _("Saving %s"), stripped.c_str());
    else
      std::snprintf (buf, sizeof(buf), _("Reading %s"), stripped.c_str());
    return std::string (buf);
  }
}

TaskArticle :: TaskArticle (const ServerRank          & server_rank,
                            const GroupServer         & group_server,
                            const Article             & article,
                            ArticleCache              & cache,
                            ArticleRead               & read,
                            Task::Listener            * listener,
                            const StringView          & save_path):
  Task (save_path.empty() ? "BODIES" : "SAVE", get_description (article, !save_path.empty())),
  _save_path (save_path.str, save_path.len),
  _server_rank (server_rank),
  _cache (cache),
  _read (read),
  _article (article),
  _time_posted (article.time_posted),
  _finished_proc_has_run (false),
  _save_when_finished (!save_path.empty())
{
  cache.reserve (article.parts);

  if (listener != 0)
    add_listener (listener);

  // build a list of all the parts we need to download.
  // also calculate need_bytes and all_bytes for our Progress status.

  quarks_t groups;
  foreach_const (Xref, article.xref, it)
    groups.insert (it->group);
  quarks_t servers;
  foreach_const (quarks_t, groups, it) {
    quarks_t tmp;
    group_server.group_get_servers (*it, tmp);
    servers.insert (tmp.begin(), tmp.end());
  }

  unsigned long need_bytes(0), all_bytes(0);
  foreach_const (Article::parts_t, article.parts, it)
  {
    all_bytes += it->bytes;
    if (!cache.contains (it->message_id))
    {
      need_bytes += it->bytes;

      Needed n;
      n.part = *it;

      // if we can keep the article-number from the main xref, do so.
      // otherwise plug in `0' as a null article-number and we'll use
      // `ARTICLE message-id' instead when talking to the server.
      foreach_const (quarks_t, servers, sit)
        foreach_const (quarks_t, groups, git)
          n.xref.insert (*sit, *git, it->message_id==article.message_id.to_string() ? article.xref.find_number(*sit,*git) : 0);

      _needed.push_back (n);
    }
  }

  // initialize our progress status...
  init_steps (all_bytes);
  set_step (all_bytes - need_bytes);
  if (save_path.empty())
    set_status (article.subject.c_str());
  else
    set_status_va (_("Saving %s"), article.subject.c_str());
 
  update_work ();
}

TaskArticle :: ~TaskArticle ()
{
  _cache.release (_article.parts);
}

void
TaskArticle :: update_work ()
{
  // which servers could we use right now?
  int working (0);
  quarks_t servers;
  foreach (needed_t, _needed, nit) {
    Needed& n (*nit);
    if (n.nntp)
      ++working;
    else {
      quarks_t tmpservers;
      while (!n.xref.empty() && tmpservers.empty()) {
        foreach_const (Xref, n.xref, xit)
          if (_server_rank.get_server_rank(xit->server) <= n.rank)
            tmpservers.insert (xit->server);
        if (tmpservers.empty())
          ++n.rank;
      }
      servers.insert (tmpservers.begin(), tmpservers.end());
    }
  }

  if (!servers.empty())
    _state.set_need_nntp (servers);
  else if (working)
    _state.set_working ();
  else {
    _state.set_completed ();
    set_finished (OK);
  }

  if (_state._work == COMPLETED && !_finished_proc_has_run) {
    _finished_proc_has_run = true;
    on_finished ();
  }
}

/***
****
***/

void
TaskArticle :: use_nntp (NNTP * nntp)
{
  // find which part, if any, can use this nntp
  Needed * needed (0);
  for (needed_t::iterator it(_needed.begin()), end(_needed.end()); !needed && it!=end; ++it)
    if (it->nntp==0 && it->xref.has_server(nntp->_server) && (it->rank <= _server_rank.get_server_rank (nntp->_server)))
      needed = &*it;

  if (!needed)
  {
    // couldn't find any work for this nntp...
    // maybe our needs-work list is out of date, so call update_work()
    // before giving the nntp back.
    update_work ();
    check_in (nntp, true);
  }
  else
  {
    needed->nntp = nntp;
    needed->buf.clear ();
    needed->buf.reserve (needed->part.bytes);

    Quark group;
    unsigned long number (0ul);
    needed->xref.find (nntp->_server, group, number);
    if (number)
      nntp->article (group, number, this);
    else
      nntp->article (group, needed->part.message_id.c_str(), this);
    update_work ();
  }
}

/***
****
***/

void
TaskArticle :: on_nntp_line  (NNTP               * nntp,
                             const StringView   & line)
{
  // FIXME: ugh, this is called for _every line_...
  Needed * needed (0);
  foreach (needed_t, _needed, it) {
    if (it->nntp == nntp) {
      needed = &*it;
      break;
    }
  }
  assert (needed);

  Needed::buf_t& buf (needed->buf);
  buf.insert (buf.end(), line.begin(), line.end());
  buf.insert (buf.end(), '\n');
  increment_step (line.len);
}

void
TaskArticle :: on_nntp_done  (NNTP    * nntp,
                              Health    health)
{
  // find the Needed using this nntp...
  needed_t::iterator it;
  for (it=_needed.begin(); it!=_needed.end(); ++it)
    if (it->nntp == nntp)
      break;
  g_assert (it != _needed.end());

  //std::cerr << LINE_ID << ' ' << it->part.message_id << " from " << nntp->_server << ": " << (health==OK ? "yes" : "no") << std::endl;

  if (health == OK) { // if download succeeded, save it in the cache
    const StringView view (&it->buf.front(), it->buf.size());
    _cache.add (it->part.message_id, view);
    ++_stats[nntp->_server];
  }

  if (health == FAIL) // if server doesn't have that article...
    it->xref.remove_server (nntp->_server);

  // if we got it or still have other options, we're okay
  _state.set_health (health==FAIL && it->xref.empty() ? FAIL : OK);

  if (health==OK)
    _needed.erase (it);
  else
    it->nntp = 0;

  update_work ();
  check_in (nntp, health!=RETRY);
}

namespace
{
  void uu_log (void* unused, char* message, int severity)
  {
     // std::cerr << LINE_ID << " severity " << severity << " message " << message << std::endl;
    if (severity == UUMSG_PANIC || severity==UUMSG_FATAL || severity==UUMSG_ERROR)
      Log :: add_err (message);
    else if (severity == UUMSG_WARNING || severity==UUMSG_NOTE)
      Log :: add_info (message);
  }
}

void
TaskArticle :: on_finished ()
{
#if 0
  // this is mostly interesting for making sure multiserver is distributing the load right...
  // so this log message will be removed before the stable release.
  // That's why it's not marked for translation
  if (_stats.size() > 1) {
    std::ostringstream o;
    unsigned int total (0);
    foreach_const (stats_t, _stats, it) total += it->second;
    o << "Fetched " << _article.subject << " in " << total << " parts: ";
    foreach_const (stats_t, _stats, it)
      o << it->second << " from " << it->first << " ";
    Log::add_info (o.str().c_str());
  }
#endif

  if (_save_when_finished)
  {
#if 0
    ArticleCache :: mid_sequence_t mids;
    foreach_const (Article::parts_t, _article.parts, it)
      mids.push_back (it->message_id);

    GMimeMessage * mm = _cache.get_message (_article.parts);
    if (mm) {
      Decode :: decode (_save_path,
                        Quark(), // FIXME: fallback filename,
                        mm,
                        *this);
      _read.mark_read (_article);
      g_object_unref (G_OBJECT (mm));
    }
#else

    // initialize uu ... this will need to be moved up s.t.
    // article reading can use it too.
    ArticleCache :: strings_t filenames = _cache.get_filenames (_article.parts);

    // decode
    int res;
    if (((res = UUInitialize())) != UURET_OK)
      Log::add_err (_("Error initializing uulib"));
    else
    {
      UUSetMsgCallback (NULL, uu_log);
      UUSetOption (UUOPT_DESPERATE, 1, NULL); // keep incompletes; they're par2 fodder
      UUSetOption (UUOPT_FAST, 1, NULL); // yes, each input file is exactly one newsgroup posting

      int i (0);
      foreach_const (ArticleCache::strings_t, filenames, it) {
        //std::cerr << LINE_ID << " loading " << *it << " as part " << (i+1) << std::endl;
        if ((res = UULoadFileWithPartNo (const_cast<char*>(it->c_str()), 0, 0, ++i)) != UURET_OK)
          Log::add_err_va (_("Couldn't load \"%s\": %s"), it->c_str(),
            (res==UURET_IOERR) ?  file::pan_strerror (UUGetOption (UUOPT_ERRNO, NULL, NULL, 0)) : UUstrerror(res));
      }

      i = 0;
      uulist * item;
      while ((item = UUGetFileListItem (i++)))
      {
        if ((item->state & UUFILE_OK) == 0)
          continue;

        char * fname (0);
        if (_save_path.empty())
          fname = g_strdup (item->filename);
        else {
          file :: ensure_dir_exists (_save_path);
          fname = g_build_filename (_save_path.c_str(), item->filename, NULL);
        }

        if ((res = UUDecodeFile (item, fname)) != UURET_OK)
          Log::add_err_va (_("Error decoding \"%s\": %s"),
                 fname, (res==UURET_IOERR) ?
                 file::pan_strerror (UUGetOption (UUOPT_ERRNO, NULL, NULL, 0)) :
                 UUstrerror(res));
        else
          Log::add_info_va (_("Saved \"%s\""), fname);
        g_free (fname);
      }
    }
    UUCleanUp ();
    
#endif
  }
}

unsigned long
TaskArticle :: get_bytes_remaining () const
{
  unsigned long bytes (0);
  foreach_const (needed_t, _needed, it) // parts not fetched yet...
    bytes += (it->part.bytes - it->buf.size());
  return bytes;
}
